# repobot/commands/base.py
'''The base command.'''

from docopt import docopt

class Base(object):
    '''A base command.'''

    def __init__(self, options, *args, **kwargs):
        self.options = options
        self.args = args
        self.kwargs = kwargs

    def run(self):
        raise NotImplementedError('The run() command was not implemented for this subclass')

    def checkHelp(self, doc):
        if '--help' in self.options:
            docopt(doc, argv=['--help'])




# repobot/commands/hello.py
'''
rbot ls - LIST REPOS

Usage:
    rbot ls [--params <str>] [--limit <number>] [--all]

Description:
    List all repos user has explicit push, pull, or admin access to (via /user/repos)
    Rbot prints list of repos on owner/name format (eg. nodejs/node), seperated on each line.
    Only prints the first 30 repos found by default, unless --all or --limit is specified.

Options:
    --params <str>      space-seperated string of key=value params, as specified in the github API
                        see https://developer.github.com/v3/repos/#list-your-repositories for details

    --limit <number>    limits the amount of entries to <number>. If a number is given thats larger
                        than the number of entries, rbot simply returs all entries. Overrules --all

    --all               List all entries. If --limit is set, this option is ignored.
                        WARNING: Will likely initiate multiple requests, depending on number of repos
                        the user has access to.
'''
import sys
from .base import Base
from json import dumps
from .utils import set_token
import json
import re
from colorama import init, Fore, Style
import requests

class Ls(Base):
    '''say hello world'''

    @set_token
    def run(self, basicauth):

        limit = int(self.options['--limit'] or -1)
        page = 1        
        last_page = self.getlastpage(basicauth=basicauth) if self.options['--limit'] or self.options['--all'] else 1 # only one page by default
        for p in range(1, last_page+1):
            repo_res = self.getreposatpage(p, basicauth=basicauth)
            for name in repo_res['repos']:
                print(name)
                limit -= 1
                # check if limit is -1 to make limit inclusive
                # don't check with less than as -1 starting is limitless (becomes -2 on first check and never breaks)
                if limit == -1:
                    return

    def getlastpage(self, *, basicauth):
        headers = requests.head('https://api.github.com/user/repos', auth=basicauth).headers
        last_page = re.compile(r'page=(\d+)>;\srel="last"').findall(headers['Link'])[0]
        return int(last_page)

    #TODO: @set_token doesnt work here. Figure out why   
    def getreposatpage(self, page, *, basicauth):
        """ """
        params = addpageparam(parseparams(self.options['--params']), page)
        res = requests.get('https://api.github.com/user/repos' + params, auth=basicauth)
        checkforerrors(res)
        rel = re.compile(r'rel="(\w+)"').findall(res.headers['Link'])[0]
        return {'repos': self.extractreponames(res.json()), 'rel': rel,}
        
        
    def extractreponames(self, data: list):
        def mapreposforprettyprint(obj):
            permissions = obj['permissions']
            if permissions['admin'] is False and permissions['push'] is False and permissions['pull'] is True:
                readonly = read = Fore.YELLOW + 'readonly' + Style.RESET_ALL
            else:
                readonly = ''
            
            priv = Fore.RED + 'private ' + Style.RESET_ALL if obj['private'] else ' '
        
            LENGTH = 24
            return obj['full_name'] + (' ' * (24-len(obj['full_name']))) + ' ' + priv + readonly 
        return map(mapreposforprettyprint, data)


def checkforerrors(res:'Response'):
    
    if res.status_code > 399:
        print(Fore.RED + 'Couldn\'t process request' + Style.RESET_ALL)
        print(json.dumps(res.json(), indent=2))
        sys.exit(1)


def parseparams(params):
    """Parses space seperated params for query string (including the starting `?`)
    Returns empty string if no params are given"""
    
    if params is None:
        return ''
    
    return '?' + re.sub(r'\s+', '&', params)

def addpageparam(params, page):
    if page == 1:
        return params
    if len(params):
        return params + '&page=' + str(page)
    return '?page=' + str(page)



# repobot/commands/hello.py
'''Test hello command'''

from .base import Base
from json import dumps

from .utils import set_token
import requests

class Hello(Base):
    '''say hello world'''

    @set_token
    def run(self, basicauth):
        
        r = requests.get('https://api.github.com/user', auth=basicauth)
        print('the re was', r)
        print('Hello, world!')
        print('You suplied the following optons:', dumps(self.options, indent=2, sort_keys=True))



"""
repobot

Usage:
    rbot login
    rbot new [<repo_name>] [-DC] [--private --clone --org=<org_name>]
    rbot ls [--params=<param_str>] [--limit <number>|--all]
    rbot pr [<repo_name>] [<branch]
    rbot hello [<world>] [--name=<yours>]

Options:
    -D                Use all defaluts.
    --clone           Automatically clone the created repo.
    --org=<org_name>  Create in the <org_name> that you're a member of.  
    --limit <number>  For getting lists, shows only the number of entries specified. Overrules --all
    --all             For getting lists, inclued all entries. Is ignored when --limit is set
    --help            Show additional descriptions for any sub commands.

Examples
    Create a new repository named foo
        rbot new foo

    Create a new private repository named foo and automatically clone it
        rbot new --private -C foo

    Create a new repository named foo in orgization xyzorg, using system defaults
        rbot new --org=xyzorg -D foo
"""

from inspect import getmembers, isclass

from docopt import docopt

from repobot import __version__ as VERSION
import sys

def main():
    """Main CLI entrypoint."""
    import repobot.commands as commands
    if '--help' in sys.argv:
        sys.argv.remove('--help')
        options = docopt(__doc__, help=False, version=VERSION)
        options['--help'] = True
    else:
        options = docopt(__doc__, help=False, version=VERSION)
    # Here we'll try to dynamically match the command the user is trying to run
    # with a pre-defined command class we've already created.
    for (k, v) in options.items():
        if hasattr(commands, k) and v:
            module = getattr(commands, k)
            rcommands = getmembers(module, isclass)
            command = [command[1] for command in rcommands if command[0] != 'Base'][0]
            command = command(options)
            command.run()

if __name__ == '__main__':
    main()



# new.py
'''
Create a new repository

Usage:
    rbot new
    rbot new <repo_name> [-DC] [--private --clone --org=<org_name>]
    
'''
import json
import sys
import requests
import os
from os import path

from ..base import Base
from ..utils import set_token, cinput, yn_input, allowescape, checkshellcommand

from docopt import docopt

class New(Base):

    @set_token
    def run(self, basicauth):
        self.checkHelp(__doc__)

        name = self.getname()

        description = self.getdescription()
        isprivate = self.getprivateoption()
        hasreadme = self.getreadmeoption()

        data = {'name': name,
                'description': description,
                'private': isprivate,
                'auto_init': hasreadme,}
        POST_URL = 'https://api.github.com/orgs/%s/repos' % (self.options['--org']) \
                    if self.options['--org'] \
                    else 'https://api.github.com/user/repos'
        
        res = requests.post(POST_URL, auth=basicauth, json=data)
        
        if res.status_code == 201:
            resdata = res.json()
            print('Successfully created at ' + resdata['clone_url'])
            self.cloneprompt(resdata['clone_url'])
        else:
            print('Couldn\'t create repo')
            print(json.dumps(res.json(), indent=2))

    @allowescape
    def getname(self):
        if self.options['<repo_name>'] is not None:
            return self.options['<repo_name>']
        return cinput('Repo name: ',
                      expression=r'^[a-z|A-Z|0-9|\-\_]*$',
                      error_message='Invalid name - use [a-z|A-Z|0-9|-_] characters only.')

    @allowescape
    def getdescription(self):
        if self.options['-D'] is not False:
            return ''
        return input('Description (optional): ')

    @allowescape
    def getprivateoption(self) -> bool:
        if self.options['-D']:
            return 'false'
        if self.options['--private']:
            return 'true'
        return yn_input('Private Repository? ', default=False)

    @allowescape
    def getreadmeoption(self) -> bool:
        if self.options['-D'] is not False:
            return 'false'
        return yn_input('Initialize with a README? ', default=False)

    @allowescape
    @checkshellcommand('git')
    def cloneprompt(self, cloneurl):
        if self.options['-C'] or yn_input('Clone into current working directory now?', default=False):
            return os.system(path.dirname(path.abspath(__file__)) + '/git-clone.sh ' + cloneurl)



# repobot/commands/utils.py
'''utility functions and wrappers'''

import sys
from functools import wraps
from requests.auth import HTTPBasicAuth
import keyring
import requests
import os

def set_token(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        username = keyring.get_password('repobot', 'username')
        password = keyring.get_password('repobot', 'password')

        if password is None:
            print('Set credentials first with `repobot login`')
            sys.exit(0)

        r = requests.get('https://api.github.com/user', auth=HTTPBasicAuth(username, password))
        if r.status_code == 401:
            print('Invalid Auth credentials. Please log in again')
            sys.exit(0)

        func(*args, basicauth=HTTPBasicAuth(username, password), **kwargs)


    return wrapper


import re

def cinput(*args, expression='', error_message='Invalid'):
    while True:
        i = input(*args)
        if re.search(expression, i):
            return i
        else:
            print(error_message)

def yn_input(prompt='', default=None):
    '''wraps input command to get a yes or no response'''
    while True:
        if default is True:
            res = input(prompt + '[Y/n] : ')
            if re.search('no?', res, flags=2):
                return False
            return True
        else:
            res = input(prompt + '[y/N] : ')
            if re.search('ye?s?', res, flags=2):
                return True
            return False

def allowescape(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            print('\nBye')
            sys.exit()
    return wrapper

def checkshellcommand(command, quiet=True):
    """
    Only runs the decorated function if the given shell commands is installed
    quiet - Silently omits running the command if true, otherwise informs user to install the required command
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if os.popen('which %s' % command).read():
                return func(*args, **kwargs)            
            if not quiet:
                print('WARN: Shell command %s was needed but not found. Some things may not work' %s)
        return wrapper
    return decorator



"""Packaging settings."""


from codecs import open
from os.path import abspath, dirname, join
from subprocess import call

from setuptools import Command, find_packages, setup

from repobot import __version__


this_dir = abspath(dirname(__file__))


with open('requirements.txt') as f:
            required = f.read().splitlines()

with open(join(this_dir, 'README.rst'), encoding='utf-8') as file:
    long_description = file.read()


class RunTests(Command):
    """Run all tests."""
    description = 'run tests'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run all tests!"""
        errno = call(['py.test', '--cov=skele', '--cov-report=term-missing'])
        raise SystemExit(errno)

setup(
    name = 'repobot',
    version = __version__,
    description = 'Programatic github repo automation',
    long_description = long_description,
    url = 'https://github.com/nickpalenchar/repobot',
    author = 'Nick Palenchar',
    author_email = '',
    license = 'MIT',
    classifiers = [ 'Development Status :: 3 - Alpha',
    ],
    keywords = 'cli',
    packages = find_packages(exclude=['docs', 'tests*']), # prevents irrelevent files from being added to package
    package_data = {'': ['*.sh', '*.bash']},
    install_requires = required,
    extras_require = {
        'test': ['coverage', 'pytest', 'pytest-cov'],
    },
    entry_points = {
        'console_scripts': [
            'rbot=repobot.cli:main',
            'repobot=repobot.cli:main'
        ],
    },
    cmdclass = {'test': RunTests},
)



# repobot/commands/login.py
'''
login - configue credentials with system keychain.

Usage:
    rbot login

Uses the system keychain to store login credentials.
You may be asked to give Python permission to access the system keychain.
Select `Always' to prevent further popups.'''

from .base import Base

import keyring
import requests
import sys
import re
import getpass

class Login(Base):
    '''login class'''

    def run(self):
        self.checkHelp(__doc__)

        username, password = self.getLoginDetails()

        res = requests.get('https://api.github.com/user', auth=requests.auth.HTTPBasicAuth(username, password))

        if res.status_code != 200:
            print('Invalid login request.')
            return sys.exit(1)
        
        keyring.set_password('repobot', 'username', username)
        keyring.set_password('repobot', 'password', password)
        print('Successfully authenticated')        

    def getLoginDetails(self) -> tuple:
        """Prompts the user for a github login, terminates on KeyboardInterrupt"""
        try:
            username = input('Guthub username: ')
            password = getpass.getpass()
        except KeyboardInterrupt:
            print('\nok then -_-')
            sys.exit()
        return (username, password)
